/**
 * @file    PAD4_DM.cpp
 * @brief   <SHORT DESCRIPTION>
 * @author  Thomas Wiemken, eQ-3 Entwicklung GmbH 
 * <OPTIONAL: LONG DESCRIPTION>
**/
/*-----------------------------------------------------------------------------
 * Copyright 2012 eQ-3 Entwicklung GmbH
 * All Rights Reserved.
 *
 * The information contained in this file is confidential property of "eQ-3
 * Entwicklung GmbH". The use copying, transfer or disclosure is prohibited
 * except by written agreement with "eQ-3 Entwicklung GmbH".
 *---------------------------------------------------------------------------*/
 
/** @addtogroup PAD4_Dot_Matrix_Hardware
 *  @{
**/
/*---------------------------------------------------------------------------*/


/**
 *  Includes ----------------------------------------------------------------- 
**/
#include "PAD4_DM.h"


/**********************************************************
Konstruktor
**********************************************************/
PAD4_DM::PAD4_DM()
{}


/**
    Functions ----------------------------------------------------------------
**/


 /** 
 * \brief Initialization of the display module object
 * 
 * Initialization of the display module object with the standard pins  
 * 
 * \param u8_number_of_modules Number of used modules 
 * 
 * \return none
 * 
**/
void PAD4_DM::begin(unsigned char u8_number_of_modules)
{
  this->u8_num_of_modules = u8_number_of_modules;
  this->u8_num_of_multiplex_cycles = NUMBER_OF_MULTIPLEX_CYCLES;
  this->u8_num_of_bits_per_module = NUMBER_OF_BITS_PER_MODULE;
  this->init_standard();
  
  /*
   * Dynamic memory allocation of a two-dimensional array, with later use of the two-dimensional array notation to access this structure.
   * array_name[a][b]
   * https://stackoverflow.com/questions/917783/how-do-i-work-with-dynamic-multi-dimensional-arrays-in-c
   */
  this->u32_display_buffer = (unsigned long**)calloc((this->u8_num_of_modules), sizeof(unsigned long*));
  
  for( unsigned char u8_module_index = 0; u8_module_index < this->u8_num_of_modules; ++u8_module_index )
  {
    this->u32_display_buffer[u8_module_index] = (unsigned long*)calloc((this->u8_num_of_multiplex_cycles), sizeof(unsigned long));
  }
}


 /** 
 * \brief Initialization of the display module object
 * 
 * Initialization of the display module object with the user-defined pins
 * 
 * \param u8_number_of_modules Number of used modules 
 * \param u8_ser_dat Pin number of the serial data pin
 * \param u8_ser_clk Pin number of the serial clock pin 
 * \param u8_ser_clr Pin number of the serial clear pin
 * \param u8_reg_clk Pin number of the register clock pin
 * \param u8_out_en Pin number of the output enable pin
 * 
 * \return none
 * 
**/
void PAD4_DM::begin(unsigned char u8_number_of_modules, unsigned char u8_ser_dat, unsigned char u8_ser_clk, unsigned char u8_ser_clr, unsigned char u8_reg_clk, unsigned char u8_out_en)
{
  this->u8_num_of_modules = u8_number_of_modules;
  this->u8_num_of_multiplex_cycles = NUMBER_OF_MULTIPLEX_CYCLES;
  this->u8_num_of_bits_per_module = NUMBER_OF_BITS_PER_MODULE;
  this->init_individual( u8_ser_dat, u8_ser_clk, u8_ser_clr, u8_reg_clk, u8_out_en );
  
  /*
   * Dynamic memory allocation of a two-dimensional array, with later use of the two-dimensional array notation to access this structure.
   * array_name[a][b]
   * https://stackoverflow.com/questions/917783/how-do-i-work-with-dynamic-multi-dimensional-arrays-in-c
   */
  this->u32_display_buffer = (unsigned long**)calloc((this->u8_num_of_modules), sizeof(unsigned long*));
  
  for( unsigned char u8_module_index = 0; u8_module_index < this->u8_num_of_modules; ++u8_module_index )
  {
    this->u32_display_buffer[u8_module_index] = (unsigned long*)calloc((this->u8_num_of_multiplex_cycles), sizeof(unsigned long));
  }
}


 /** 
 * \brief Hardware Initialization
 * 
 * Hardware initialization via the standard pins.
 * 
 * \return none
 * 
**/ 
void PAD4_DM::init_standard( void )
{
  pinMode(this->u8_pin_layout[SERIAL_DATA_PIN], OUTPUT);      // sets the digital pin SERIAL_DATA as output
  pinMode(this->u8_pin_layout[SERIAL_CLOCK_PIN], OUTPUT);     // sets the digital pin SERIAL_CLOCK as output
  pinMode(this->u8_pin_layout[SERIAL_CLEAR_PIN], OUTPUT);     // sets the digital pin SERIAL_CLEAR as output
  pinMode(this->u8_pin_layout[REGISTER_CLOCK_PIN], OUTPUT);   // sets the digital pin REGISTER_CLOCK as output
  pinMode(this->u8_pin_layout[OUTPUT_ENABLE_PIN], OUTPUT);    // sets the digital pin OUTPUT_ENABLE as output
  
  digitalWrite(this->u8_pin_layout[SERIAL_DATA_PIN], LOW);    // sets the digital pin SERIAL_DATA to low
  digitalWrite(this->u8_pin_layout[SERIAL_CLOCK_PIN], LOW);   // sets the digital pin SERIAL_CLOCK to low
  digitalWrite(this->u8_pin_layout[SERIAL_CLEAR_PIN], LOW);   // sets the digital pin SERIAL_CLEAR to low
  delay(1);                                                   // waits for 1 miliseconds
  digitalWrite(this->u8_pin_layout[SERIAL_CLEAR_PIN], HIGH);  // sets the digital pin SERIAL_CLEAR to high
  digitalWrite(this->u8_pin_layout[REGISTER_CLOCK_PIN], LOW); // sets the digital pin REGISTER_CLOCK to low
  delay(1);                                                   // waits for 1 miliseconds
  digitalWrite(this->u8_pin_layout[REGISTER_CLOCK_PIN], LOW); // sets the digital pin REGISTER_CLOCK to high
  digitalWrite(this->u8_pin_layout[OUTPUT_ENABLE_PIN], LOW);  // sets the digital pin OUTPUT_ENABLE to low
}


 /** 
 * \brief Hardware Initialization
 * 
 * Hardware initialization via the user-defined pins. 
 *  
 * \param u8_ser_dat Pin number of the serial data pin
 * \param u8_ser_clk Pin number of the serial clock pin 
 * \param u8_ser_clr Pin number of the serial clear pin
 * \param u8_reg_clk Pin number of the register clock pin
 * \param u8_out_en Pin number of the output enable pin
 * 
 * \return none
 * 
**/ 
void PAD4_DM::init_individual( unsigned char u8_ser_dat, unsigned char u8_ser_clk, unsigned char u8_ser_clr, unsigned char u8_reg_clk, unsigned char u8_out_en )
{
  this->u8_pin_layout[SERIAL_DATA_PIN] = u8_ser_dat;
  this->u8_pin_layout[SERIAL_CLOCK_PIN] = u8_ser_clk;
  this->u8_pin_layout[SERIAL_CLEAR_PIN] = u8_ser_clr;
  this->u8_pin_layout[REGISTER_CLOCK_PIN] = u8_reg_clk;
  this->u8_pin_layout[OUTPUT_ENABLE_PIN] = u8_out_en;
  
  pinMode(this->u8_pin_layout[SERIAL_DATA_PIN], OUTPUT);      // sets the digital pin SERIAL_DATA as output
  pinMode(this->u8_pin_layout[SERIAL_CLOCK_PIN], OUTPUT);     // sets the digital pin SERIAL_CLOCK as output
  pinMode(this->u8_pin_layout[SERIAL_CLEAR_PIN], OUTPUT);     // sets the digital pin SERIAL_CLEAR as output
  pinMode(this->u8_pin_layout[REGISTER_CLOCK_PIN], OUTPUT);   // sets the digital pin REGISTER_CLOCK as output
  pinMode(this->u8_pin_layout[OUTPUT_ENABLE_PIN], OUTPUT);    // sets the digital pin OUTPUT_ENABLE as output
  
  digitalWrite(this->u8_pin_layout[SERIAL_DATA_PIN], LOW);    // sets the digital pin SERIAL_DATA to low
  digitalWrite(this->u8_pin_layout[SERIAL_CLOCK_PIN], LOW);   // sets the digital pin SERIAL_CLOCK to low
  digitalWrite(this->u8_pin_layout[SERIAL_CLEAR_PIN], LOW);   // sets the digital pin SERIAL_CLEAR to low
  delay(1);                                                   // waits for 1 miliseconds
  digitalWrite(this->u8_pin_layout[SERIAL_CLEAR_PIN], HIGH);  // sets the digital pin SERIAL_CLEAR to high
  digitalWrite(this->u8_pin_layout[REGISTER_CLOCK_PIN], LOW); // sets the digital pin REGISTER_CLOCK to low
  delay(1);                                                   // waits for 1 miliseconds
  digitalWrite(this->u8_pin_layout[REGISTER_CLOCK_PIN], LOW); // sets the digital pin REGISTER_CLOCK to high
  digitalWrite(this->u8_pin_layout[OUTPUT_ENABLE_PIN], LOW);  // sets the digital pin OUTPUT_ENABLE to low
}


 /** 
 * \brief Get number of modules
 * 
 * \return The number of used display moduls
 * 
**/
unsigned char PAD4_DM::get_number_of_modules( void )
{
  return( this->u8_num_of_modules);
}


 /** 
 * \brief Get number of available digits
 * 
 * \return The number of used available digits
 * 
**/
unsigned char PAD4_DM::get_number_of_digits( void )
{
  return( (this->u8_num_of_modules * NUMBER_OF_DIGITS_PER_MODULE) );
}


 /** 
 * \brief Set up display buffer
 * 
 * Fills the display buffer with the data from the text string array
 * 
 * \param *s8_text Pointer to the text string array 
 * 
 * \return none
 * 
**/ 
void PAD4_DM::set_display_buffer( char *s8_text )
{
  unsigned long u32_temp_value = 0;
  unsigned int u16_bit_mask = 0x2000;
  unsigned int u16_temp_value = 0;
  
  for( unsigned char u8_module_index = 0; u8_module_index < this->u8_num_of_modules; u8_module_index++ )
  {
    for( unsigned char u8_multiplex_cycle = 0; u8_multiplex_cycle < this->u8_num_of_multiplex_cycles; u8_multiplex_cycle++ )
    {
      u16_temp_value = pgm_read_byte(&font_5x7[s8_text[((2*u8_module_index)+1)]-0x20][u8_multiplex_cycle]);  // Get data for the first character of the module
      u16_temp_value *= 128;  // Shift the bits 7 times left
      u16_temp_value += pgm_read_byte(&font_5x7[s8_text[((2*u8_module_index)+0)]-0x20][u8_multiplex_cycle]);  // Get data for the second character of the module
      
      // Set up the data bit variable
      u32_temp_value = 0;
      // Set up the bit mask variable 
      u16_bit_mask = 0x2000;
      
      for( unsigned char index = 0; index < 14; index++ )
      {
        
        if( u16_temp_value & u16_bit_mask )
        {
          u32_temp_value |= (1<<index);
        }
        
        // Shift the bit mask right
        u16_bit_mask /= 2;
        
      }
      
      // Shift the data bits 5 times left
      u32_temp_value *= 32; // 0b0000000000XXXXXXXXXXXXXX --> 0b00000XXXXXXXXXXXXXX00000
    
      switch( u8_multiplex_cycle )
      {
        case 0:
        u32_temp_value |= 0b011110000000000000001111;
        break;
        
        case 1:
        u32_temp_value |= 0b101110000000000000010111;
        break;
        
        case 2:
        u32_temp_value |= 0b110110000000000000011011;
        break;
        
        case 3:
        u32_temp_value |= 0b111010000000000000011101;
        break;
        
        case 4:
        u32_temp_value |= 0b111100000000000000011110;
        break;
        
        default:
        u32_temp_value |= 0b111110000000000000011111;
        break;
      }
      // copy the created data bits to the display buffer
      this->u32_display_buffer[u8_module_index][u8_multiplex_cycle] = u32_temp_value;
    }
  }
}


 /** 
 * \brief Display Matrix Update
 * 
 * Updates the content of the display content 
 * 
 * \return none 
 * 
**/ 
void PAD4_DM::update_matrix( void )
{
  static unsigned char u8_multiplex_cycle = 0xff;
  unsigned long u32_bit_mask = 0x00000001;
  
  // Reset the multiplex cycle number, if the end has reached
  if( u8_multiplex_cycle >= this->u8_num_of_multiplex_cycles )
  {
    u8_multiplex_cycle = 0;
  }
  
  // Clear the shift register
  digitalWrite(this->u8_pin_layout[SERIAL_CLEAR_PIN], LOW);    // sets the digital pin SERIAL_CLEAR to low
  delayMicroseconds(1);
  digitalWrite(this->u8_pin_layout[SERIAL_CLEAR_PIN], HIGH);   // sets the digital pin SERIAL_CLEAR to high

  // Fill the shift register of each used module with the data from the display buffer
  for( unsigned char u8_module_index = this->u8_num_of_modules; u8_module_index > 0; u8_module_index-- )
  {
    // Set bit mask to start value
    u32_bit_mask = 0x00000001;
    
    // Send 24 bits per multiplex cycle
    for( unsigned char u8_index = 0; u8_index < this->u8_num_of_bits_per_module; u8_index++ )
    {
      // Set the data pin 
      if(this->u32_display_buffer[(u8_module_index-1)][u8_multiplex_cycle] & u32_bit_mask)
      {
        digitalWrite(this->u8_pin_layout[SERIAL_DATA_PIN], HIGH);
      }
      else
      {
        digitalWrite(this->u8_pin_layout[SERIAL_DATA_PIN], LOW);
      }
      
      // Get the data pin value into the shift register
      digitalWrite(this->u8_pin_layout[SERIAL_CLOCK_PIN], HIGH);
      delayMicroseconds(1);
      digitalWrite(this->u8_pin_layout[SERIAL_CLOCK_PIN], LOW);
      delayMicroseconds(1);
  
      // Shift the bit mask left
      u32_bit_mask *=2;
    }
  }
  
  // Reset data pin
  digitalWrite(this->u8_pin_layout[SERIAL_DATA_PIN], LOW);
  
  // Latch the shift register to the output register
  digitalWrite(this->u8_pin_layout[REGISTER_CLOCK_PIN], HIGH);  // sets the digital pin REGISTER_CLOCK to low
  delayMicroseconds(1);
  digitalWrite(this->u8_pin_layout[REGISTER_CLOCK_PIN], LOW); // sets the digital pin REGISTER_CLOCK to high
  
  // Increase number of multiplex cycle
  u8_multiplex_cycle++;
}
